'''
Copyright (C) 2015 by the University of Southern California

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License,
version 2, as published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License along
with this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.

rdns.py
-
A script to do reverse DNS lookups of IPv4.

RANGE = RANGE_START-RANGE_END
    Determines where we probe and by default we will probe from 0 to 255.
    Ex: 1 probes 1.0.0.0 to 1.255.255.255 & 1-5 probes 1.0.0.0 to 5.255.255.255

PROGRAM_MODE
    Dictates whether program is probing the range or only doing retries.
    0: Retry mode &  1: Normal mode

When using the -s arg to set the source ID make sure it is all one string.
    Works: usc-isi or uscisi Doesn't work: usc isi
'''

import sys,time,argparse,base64,getdns
from ip_iterator import IP_Iterator

__author__ = "Romello Goodman"
__version__ = "0.1.0"
__maintainer__ = "USC Information Sciences Institute"

# Set defaults
file = sys.stdout
program_mode = 1
timeout = 2000
query_limit = 20
query_amt = 0
source_id = 'USC-ISI'
resolver = '127.0.0.1'
range_start = '0'
range_end = '255'
current_instance = 0
total_instances = 1
ip_counter = 0
timeoutfile = 'timeout.txt'
black_list = [
    '0.0.0.0-0.255.255.255',        #0/8
    '10.0.0.0-10.255.255.255',      #10/8
    '169.254.0.0-169.254.255.255',  #169.254/16
    '172.16.0.0-172.31.255.255',    #172.16/12
    '192.0.0.0-192.0.0.255',        #192.0.2/24
    '192.168.0.0-192.168.255.255',  #192.168/16
    '224.0.0.0-224.255.255.255',    #224/4
    '240.0.0.0-240.255.255.255'     #240/4
]
stats = {
    'Found': 0,
    'Not Found': 0,
    'Timeout': 0,
    'Canceled': 0,
    'Error': 0
}
ctx = getdns.Context()
ctx.timeout = timeout
ctx.dns_root_servers = [ { 'address_data': resolver, 'address_type': 'IPv4' } ]
extensions = { "dnssec_return_validation_chain" : getdns.EXTENSION_TRUE }
ctx.resolution_type = getdns.RESOLUTION_STUB

# Add & Parse Args
parser = argparse.ArgumentParser()
parser.add_argument("range", type=str)
parser.add_argument("--current_instance", type=int)
parser.add_argument("--total_instances", type=int)
parser.add_argument("--timeoutfile", type=str)
parser.add_argument("-i", "--program_mode", type=int)
parser.add_argument("-o", "--output", type=str)
parser.add_argument("-q", "--query_limit", type=int)
parser.add_argument("-r", "--resolver", type=str)
parser.add_argument("-s", "--source_id", type=str)
parser.add_argument("-t", "--timeout", type=int)
args = parser.parse_args()

# Split the range
if len(args.range.split('-'))>1:
    range_start,range_end = args.range.split('-')
else:
    range_start,range_end = args.range,args.range
if args.current_instance:   current_instance=args.current_instance
if args.total_instances:    total_instances=args.total_instances
if args.timeoutfile:        timeoutfile=args.timeoutfile
if args.program_mode:       program_mode=args.program_mode
if args.output:             file=open(args.output,"w")
if args.query_limit:        query_limit=args.query_limit
if args.resolver:           resolver=args.resolver
if args.source_id:          source_id=args.source_id
if args.timeout:            timeout=args.timeout

def reverse_lookup(ip, extensions):
    # Concat the time the query was submitted and the ip for our userarg
    userarg = ip + '-' + str(time.time())
    try:
        ipname={'address_type':'IPv4', 'address_data':ip}
        ctx.hostname(ipname, callback='cbk', userarg=userarg, extensions=extensions)
    except getdns.error, e:
        print(str(e))

def cbk(type, result, userarg, tid):
    global query_amt
    # Split the userarg into an array: 0 is the IP and 1 is the time submitted
    userarg = userarg.split('-')
    if type == getdns.CALLBACK_COMPLETE:
        status,host,rrsig,authority = result.status,'-','-','-'
        if result.validation_chain:
            rrsig = base64.b64encode(result.validation_chain[len(result.validation_chain)-1]['rdata']['signature'])
        result = result.replies_tree[0]
        if status == getdns.RESPSTATUS_GOOD:
            host = result['answer'][0]['rdata']['ptrdname']
            write_data('Found',userarg[0],host,authority,rrsig,userarg[1])
        elif status == getdns.RESPSTATUS_NO_NAME:
            if result.has_key('authority'):
                authority = (result['authority'][0]['rdata']['mname'] +
                    '/' + result['authority'][0]['rdata']['rname'] +
                    '/' + str(result['authority'][0]['rdata']['serial']))
            write_data('Not Found',userarg[0],host,authority,rrsig,userarg[1])
        else:
            write_data('Error',userarg[0],'','','','')
    elif type == getdns.CALLBACK_CANCEL:
        write_data('Canceled',userarg[0],'','','','')
    elif type == getdns.CALLBACK_TIMEOUT:
        write_data('Timeout',userarg[0],'','','','')
    else:
        write_data('Error',userarg[0],'','','','')
    query_amt -=1

def write_data(found_status, ip, host, authority, rrsig, time_submitted):
    stats[found_status] += 1
    if(found_status == 'Found') or (found_status == 'Not Found'):
        data = ('%s\t%s\t%s\t%s\t%s\t%s\t%s\t%s\n' % (get_block(ip),ip,host,authority,rrsig,str(time.time()),str(time.time() - float(time_submitted)),source_id))
        # Block IP Answer Authority RRSIG Timestamp Duration Source
        file.write(data)
    else:
        f = open(timeoutfile,'a')
        f.write(ip + '\n')
        f.close()

def get_block(ip):
    block = ''
    octects = ip.split('.')
    for i in range(3):
        if int(octects[i]) < 10:
            block += '0' + hex(int(octects[i])).replace('0x', '')
        else:
            block += hex(int(octects[i])).replace('0x', '')
    return block

def convert_range(start, end):
    start_number = int(hex(start) + '000000', 16)
    end_number  = int(hex(end)  + 'FFFFFF', 16)
    return start_number, end_number

# Check that the ip IS NOT on the blackList and IS within the range
def allowed_value(ip):
    # check if in range
    if(ip < range_start) or (ip > range_end):
        return False
    # blackList
    if len(black_list) > 0:
        # Gets the start and end points of an ip
        for iprange in black_list:
            iprange = iprange.split('-')
            start,end = ip_to_num(iprange[0]),ip_to_num(iprange[1])
            if(ip >= start) and (ip <= end):
                return False
    return True

# Taken from Ben Schmidt's ip_utils.py at
# http://spareclockcycles.org/2010/04/13/reverse-dns-lookups-with-dnspython/
def ip_to_num(ip):
    '''Takes an IP string and returns its decimal equivalent'''
    octects = ip.split(".")
    if len(octects) == 4:
        num = sum([int(octects[i])<<((3-i)*8) for i in range(0,4)])
        if num < 0 or num > 0xFFFFFFFF: return None
        else: return num
    else:
        return None

def num_to_ip(num):
    '''Takes an decimal representation of an IP and converts to a string'''
    return ".".join([str((num>>((3-i)*8))&0xFF) for i in range(0,4)])

# Run our program depending on which mode you are in
if program_mode == 0:
    # Open the file and retry the ips that were not found
    f = open(timeoutfile, 'r')
    for line in iter(f):
        while(query_amt > query_limit):
            ctx.run()
            pass
        query_amt += 1
        reverse_lookup(line, extensions)
    f.close()
    if query_amt > 0:
        ctx.run()
else:
    # Write the header
    file.write('#fsdb F -t block\tIP\tanswer\tauthority\trrsig\ttimestamp\tduration\tsource\n')

    ip_list = IP_Iterator()
    # ip_list = IP_Iterator(32, 0x80090000, 16, 0)
    range_start,range_end = convert_range(int(range_start), int(range_end))

    # Start time
    stime = time.time()

    for ip in ip_list:
        if(allowed_value(ip)):
            ip = num_to_ip(ip)
            # Check the lower bits to see if this ip can be probed by this instance
            if((int(ip.split('.')[3]) % total_instances == current_instance)):
                while(query_amt > query_limit):
                    ctx.run()
                    pass
                query_amt += 1
                ip_counter += 1
                reverse_lookup(ip, extensions)
    if query_amt > 0:
        ctx.run()

    # Print resulting stats
    ttime = time.time() - stime
    print ("Time Taken: %s\nTotal IPs Scanned: %s\nIPs/sec: %s" % (ttime,ip_counter,ip_counter/ttime))
    for key in stats:
        print 'IP\'s ' + key + ': ' + str(stats[key])
time.sleep(timeout/1000)
